/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2007 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Anatoliy Belsky ab@php.net                                   |
  +----------------------------------------------------------------------+
*/

/* $Id: header,v 1.16.2.1.2.1 2007/01/01 19:32:09 iliaa Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/pcre/php_pcre.h"

#include "gender_definitions.h"

#include "php_gender.h"

extern int initialize_gender (TSRMLS_D);
extern int get_gender (char first_name[], int compare_mode, int country TSRMLS_DC);
extern int determine_country (char *text TSRMLS_DC);

/* class definitions */
static zend_class_entry *Gender_ce;

/**
 * The function will parse a dsn string as follows:
 * mysql://user:password@host:port/dbname
 *   [0] => mysql://user:password@host:port/dbname
 *   [1] => transport
 *   [2] => user
 *   [3] => password
 *   [4] => host
 *   [5] => port
 *   [6] => db
 * @param char *dsn
 * @return HashTAble of *chars
 * TODO this function isn't really good to parse a dsn string,
 * need to write one which would parse dsn strings pear alike
 */
static HashTable *gender_parse_dsn(const char *dsn TSRMLS_DC)
{
    pcre *re;
    const int OFFSETS_LEN = 21;
    int rc, erroffset, offsets[OFFSETS_LEN];
    const char *error;
    HashTable *ret = NULL;

    /**
     * This pattern does match the things like
     * mysql://user:password@host:port/dbname
     */
    /*                       transport               user     password        host                 port      db*/
    char *pattern = "(mysql|postgres|sqlite)://([^:]+)?:?([^@]+)?@?([[^/]|[a-zA-Z0-9-.]]*)?:?([^/]+)?/?(.*)?";

    re = pcre_compile(pattern, 0, &error, &erroffset, 0);

    if (re == NULL)
    {
            zend_throw_exception_ex(Gender_ce, 0 TSRMLS_CC, "PCRE compilation failed at offset %d: %s", &erroffset, &error);

            return ret;
    }

    rc = pcre_exec(re, pcre_study(re, 0, &error), dsn, strlen(dsn)+1, 0, PCRE_NOTEMPTY, offsets, OFFSETS_LEN);

    if(rc < 0)
    {
            pcre_free(re);

            return ret;
    } else if(rc) {

        ALLOC_HASHTABLE(ret);
        zend_hash_init(ret, rc, NULL, NULL, 0);

        int i;

        for(i = 0; i < rc; i++) {
            char *ctmp = NULL;

            ctmp = emalloc(offsets[2*i+1] - offsets[2*i] + 1);

            php_sprintf(ctmp, "%.*s", offsets[2*i+1] - offsets[2*i], dsn + offsets[2*i]);

            zend_hash_next_index_insert(ret, (void*)ctmp, strlen(ctmp)+1, NULL);

        }
    }

    return ret;
}


int gender_connect_to_source(const char *dsn TSRMLS_DC)
{
        if(!dsn) {
            zend_throw_exception(Gender_ce, "Invalid or empty dsn given", 0 TSRMLS_CC);
        
            return INTERNAL_ERROR_GENDER;
        }

        HashTable *path_info  = gender_parse_dsn(dsn TSRMLS_CC);

        /**
         * If it's a valid database dsn, try to connect,
         * if not - use streams (not yet implemented).
         */
        if(path_info) {

/*            if(zend_hash_index_find(path_info, 1, (void **)&tmp) != FAILURE) {

            }*/
/*
            for(zend_hash_internal_pointer_reset(path_info); zend_hash_has_more_elements(path_info) == SUCCESS; zend_hash_move_forward(path_info)) {
                char *item;

                if(zend_hash_get_current_data(path_info, (void **)&item) == FAILURE) {
                    php_printf("hello error\n");
                    continue;
                }

                php_printf("item: %s\n", item);
            }*/

            /* zend_hash_destroy(path_info); */
        } else {
            /**
             * dsn string was not parsed,
             * so suppose some streamable path
             */
            GENDER_G(data_mode) = DATA_MODE_ORIGINAL;
            GENDER_G(dsn) = (char*)dsn;

            if (GENDER_G(f_names) == NULL) {
                GENDER_G(f_names) = php_stream_open_wrapper(GENDER_G(dsn), "r",  STREAM_MUST_SEEK | USE_PATH, NULL);

                if (GENDER_G(f_names) == NULL) {
                    zend_throw_exception_ex(Gender_ce, 0 TSRMLS_CC, "Could not open data file '%s'", estrdup(dsn));

                    return INTERNAL_ERROR_GENDER;
                }
            }

        }

        return 0;
}


ZEND_DECLARE_MODULE_GLOBALS(gender);

void gender_globals_dtor(zend_gender_globals *gender_globals TSRMLS_DC)
{

}

void gender_globals_ctor(zend_gender_globals *gender_globals TSRMLS_DC)
{
    gender_globals->internal_mode = GENDER_DEFAULT_TRACE_MODE;
    gender_globals->f_names = NULL;
    gender_globals->line_size = 0;
    gender_globals->record_count = 0L;
    
    gender_globals->data_mode = DATA_MODE_ORIGINAL;
    gender_globals->dsn = NULL;
}

/* True global resources - no need for thread safety here */
static int le_gender;

/* {{{ client class methods[] */
zend_function_entry Gender_methods[] = {
        PHP_ME(Gender, __construct, NULL, ZEND_ACC_PUBLIC)
        PHP_ME(Gender, __destruct, NULL, ZEND_ACC_PUBLIC)
        PHP_ME(Gender, get, NULL, ZEND_ACC_PUBLIC)
        PHP_ME(Gender, connect, NULL, ZEND_ACC_PUBLIC)
        PHP_ME(Gender, trace, NULL, ZEND_ACC_PUBLIC)
    {NULL, NULL, NULL}
};
/* }}} */

ZEND_BEGIN_ARG_INFO_EX(Gender_get, 0, 0, 2)
    ZEND_ARG_INFO(0, firstname)
    ZEND_ARG_INFO(0, country)
ZEND_END_ARG_INFO()

/* {{{ gender_functions[]
 *
 * Every user visible function must have an entry in gender_functions[].
 */
zend_function_entry gender_functions[] = {
    {NULL, NULL, NULL}
};
/* }}} */

/* {{{ gender_module_entry
 */
zend_module_entry gender_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    "gender",
    gender_functions,
    PHP_MINIT(gender),
    PHP_MSHUTDOWN(gender),
    PHP_RINIT(gender),        /* Replace with NULL if there's nothing to do at request start */
    PHP_RSHUTDOWN(gender),    /* Replace with NULL if there's nothing to do at request end */
    PHP_MINFO(gender),
#if ZEND_MODULE_API_NO >= 20010901
    "0.1", /* Replace with version number for your extension */
#endif
    STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_GENDER
ZEND_GET_MODULE(gender)
#endif


/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(gender)
{

#ifdef ZTS
    ts_allocate_id(&gender_globals_id,
                   sizeof(zend_gender_globals),
                   (ts_allocate_ctor)gender_globals_ctor,
                   (ts_allocate_dtor)gender_globals_dtor);
#else
    gender_globals_ctor(&gender_globals TSRMLS_CC);
#endif

    /* initalizing classes */
    zend_class_entry ce;
        
#if USING_PHP_53
#define GENDER_GENDER_CLASS_NAME "Gender"
    INIT_CLASS_ENTRY(ce, "Gender::Gender", Gender_methods);
#else
#define GENDER_GENDER_CLASS_NAME "Gender"
    INIT_CLASS_ENTRY(ce, GENDER_GENDER_CLASS_NAME, Gender_methods);
#endif

    Gender_ce = zend_register_internal_class(&ce TSRMLS_CC);

    /* name types list */
    zend_declare_class_constant_long(Gender_ce, "IS_FEMALE", sizeof("IS_FEMALE")-1, IS_FEMALE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IS_MOSTLY_FEMALE", sizeof("IS_MOSTLY_FEMALE")-1, IS_MOSTLY_FEMALE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IS_MALE", sizeof("IS_MALE")-1, IS_MALE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IS_MOSTLY_MALE", sizeof("IS_MOSTLY_MALE")-1, IS_MOSTLY_MALE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IS_UNISEX_NAME", sizeof("IS_UNISEX_NAME")-1, IS_UNISEX_NAME TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IS_A_COUPLE", sizeof("IS_A_COUPLE")-1, IS_A_COUPLE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "NAME_NOT_FOUND", sizeof("NAME_NOT_FOUND")-1, NAME_NOT_FOUND TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ERROR_IN_NAME", sizeof("ERROR_IN_NAME")-1, ERROR_IN_NAME TSRMLS_CC);

    /* countries list */
    zend_declare_class_constant_long(Gender_ce, "ANY_COUNTRY", sizeof("ANY_COUNTRY")-1, GC_ANY_COUNTRY TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "BRITAIN", sizeof("BRITAIN")-1, GC_BRITAIN TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "IRELAND", sizeof("IRELAND")-1, GC_IRELAND TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "USA", sizeof("USA")-1, GC_USA TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "SPAIN", sizeof("SPAIN")-1, GC_SPAIN TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "PORTUGAL", sizeof("PORTUGAL")-1, GC_PORTUGAL TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ITALY", sizeof("ITALY")-1, GC_ITALY TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "MALTA", sizeof("MALTA")-1, GC_MALTA TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "FRANCE", sizeof("FRANCE")-1, GC_FRANCE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "BELGIUM", sizeof("BELGIUM")-1, GC_BELGIUM TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "LUXEMBOURG", sizeof("LUXEMBOURG")-1, GC_LUXEMBOURG TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "NETHERLANDS", sizeof("NETHERLANDS")-1, GC_NETHERLANDS TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "GERMANY", sizeof("GERMANY")-1, GC_GERMANY TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "EAST_FRISIA", sizeof("EAST_FRISIA")-1, GC_EAST_FRISIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "AUSTRIA", sizeof("AUSTRIA")-1, GC_AUSTRIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "SWISS", sizeof("SWISS")-1, GC_SWISS TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "ICELAND", sizeof("ICELAND")-1, GC_ICELAND TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "DENMARK", sizeof("DENMARK")-1, GC_DENMARK TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "NORWAY", sizeof("NORWAY")-1, GC_NORWAY TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "SWEDEN", sizeof("SWEDEN")-1, GC_SWEDEN TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "FINLAND", sizeof("FINLAND")-1, GC_FINLAND TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ESTONIA", sizeof("ESTONIA")-1, GC_ESTONIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "LATVIA", sizeof("LATVIA")-1, GC_LATVIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "LITHUANIA", sizeof("LITHUANIA")-1, GC_LITHUANIA TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "POLAND", sizeof("POLAND")-1, GC_POLAND TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "CZECH_REP", sizeof("CZECH_REP")-1, GC_CZECH_REP TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "SLOVAKIA", sizeof("SLOVAKIA")-1, GC_SLOVAKIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "HUNGARY", sizeof("HUNGARY")-1, GC_HUNGARY TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ROMANIA", sizeof("ROMANIA")-1, GC_ROMANIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "BULGARIA", sizeof("BULGARIA")-1, GC_BULGARIA TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "BOSNIA", sizeof("BOSNIA")-1, GC_BOSNIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "CROATIA", sizeof("CROATIA")-1, GC_CROATIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "KOSOVO", sizeof("KOSOVO")-1, GC_KOSOVO TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "MACEDONIA", sizeof("MACEDONIA")-1, GC_MACEDONIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "MONTENEGRO", sizeof("MONTENEGRO")-1, GC_MONTENEGRO TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "SERBIA", sizeof("SERBIA")-1, GC_SERBIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "SLOVENIA", sizeof("SLOVENIA")-1, GC_SLOVENIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ALBANIA", sizeof("ALBANIA")-1, GC_ALBANIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "GREECE", sizeof("GREECE")-1, GC_GREECE TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "RUSSIA", sizeof("RUSSIA")-1, GC_RUSSIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "BELARUS", sizeof("BELARUS")-1, GC_BELARUS TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "MOLDOVA", sizeof("MOLDOVA")-1, GC_MOLDOVA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "UKRAINE", sizeof("UKRAINE")-1, GC_UKRAINE TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ARMENIA", sizeof("ARMENIA")-1, GC_ARMENIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "AZERBAIJAN", sizeof("AZERBAIJAN")-1, GC_AZERBAIJAN TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "GEORGIA", sizeof("GEORGIA")-1, GC_GEORGIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "KAZAKH_UZBEK", sizeof("KAZAKH_UZBEK")-1, GC_KAZAKH_UZBEK TSRMLS_CC);

    zend_declare_class_constant_long(Gender_ce, "TURKEY", sizeof("TURKEY")-1, GC_TURKEY TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ARABIA", sizeof("ARABIA")-1, GC_ARABIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "ISRAEL", sizeof("ISRAEL")-1, GC_ISRAEL TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "CHINA", sizeof("CHINA")-1, GC_CHINA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "INDIA", sizeof("INDIA")-1, GC_INDIA TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "JAPAN", sizeof("JAPAN")-1, GC_JAPAN TSRMLS_CC);
    zend_declare_class_constant_long(Gender_ce, "KOREA", sizeof("KOREA")-1, GC_KOREA TSRMLS_CC);

    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(gender)
{

#ifdef ZTS
    ts_free_id(gender_globals_id);
#else
    gender_globals_dtor(&gender_globals TSRMLS_CC);
#endif
    
    return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(gender)
{
    return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request end */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(gender)
{
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(gender)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "gender support", "enabled");
    php_info_print_table_end();
}
/* }}} */


/* {{{ proto Gender Gender::Gender([string $dsn]) construct Gender optionally connecting to data source at the same time */
PHP_METHOD(Gender, __construct)
{
    int i;
    zval *dsn = NULL;
    
    if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|z", &dsn) == FAILURE) {
            return;
    }

    if(dsn) {
        convert_to_string(dsn);
        GENDER_G(dsn) = strdup(Z_STRVAL_P(dsn));
        if(gender_connect_to_source(GENDER_G(dsn) TSRMLS_CC) == INTERNAL_ERROR_GENDER) {
            RETURN_NULL();
        }
    }

    if (! (GENDER_G(internal_mode) & GENDER_INITIALIZED)) {
        i = initialize_gender(TSRMLS_C);

        if (i < 0  ||  ! (GENDER_G(internal_mode) & GENDER_INITIALIZED)) {
            zend_throw_exception(Gender_ce, "Initialization failed", 0 TSRMLS_CC);
            RETURN_NULL();
        }
    }
}
/* }}} */


/* {{{ proto void Gender::__destruct(void) */
PHP_METHOD(Gender, __destruct)
{
    cleanup_gender(TSRMLS_C);
}
/* }}} */


/* {{{ proto integer Gender::get(string $name, string $country) get gender for the name in the country */
PHP_METHOD(Gender, get)
{
    char *firstname;
    int firstname_len, gc_country = 0;
    zval *country = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|z", &firstname, &firstname_len, &country) == FAILURE) {
            return;
    }

    if(firstname_len) {
        if(country) {
            if(Z_TYPE_P(country) == IS_LONG) {
                gc_country = Z_LVAL_P(country);
            } else if(Z_TYPE_P(country) == IS_STRING) {
                gc_country = determine_country(Z_STRVAL_P(country) TSRMLS_CC);
            }
        }

        ZVAL_LONG(return_value, get_gender(firstname, 0, gc_country TSRMLS_CC));

        return;
    }

    RETURN_FALSE;
}
/* }}} */


/* {{{ proto boolean Gender::connect(string $dsn) open a data source, will support several sources*/
PHP_METHOD(Gender, connect)
{

    char *dsn, *transport, *tmp;
    int dsn_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &dsn, &dsn_len) == FAILURE) {
            RETURN_FALSE;
    }

    if(dsn_len) {
        if(gender_connect_to_source(dsn TSRMLS_CC) != INTERNAL_ERROR_GENDER) {
            RETURN_TRUE;
        }
    }

    RETURN_FALSE;

}
/* }}} */


/* {{{ proto boolean Gender::trace(void) allow trace mode */
PHP_METHOD(Gender, trace)
{
    GENDER_G(internal_mode) = TRACE_GENDER;
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
