/*

This file is subject to the GNU Lesser General Public License (LGPL)
(formerly known as GNU Library General Public Licence)
as published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

 Original program:
 Project "gender" from www.heise.de/ct, soft-link 0717182
 or  http://www.heise.de/ct/07/17/182
 Program "gender.c" and dictionary file "nam_dict.txt":
 (c) by Jrg MICHAEL, Hannover, Germany, 2007-2008

 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gender_definitions.h"

#include "php.h"
#include "php_ini.h"
#include "php_gender.h"

extern int gender_connect_to_source(const char *dsn TSRMLS_DC);

/****  data file info  ****/
struct gc_struct
   { unsigned int pos;
     unsigned int n;
     int  quality_of_statistics;
     int  weight;
     int  gc_country;
     char *country_short;
     char *country_text;
   };


/****  Notice: "country_short" is mostly (but not always)  ****/
/****          identical to international country codes    ****/
static struct gc_struct gc_data[]
= { { 30, 0, VERY_GOOD, 60, GC_BRITAIN,   "UK",   "Great Britain" },
    { 31, 0, GOOD,       4, GC_IRELAND,   "IRE",  "Ireland"     },
    { 32, 0, VERY_GOOD,150, GC_USA,       "USA",  "U.S.A."      },

    { 33, 0, VERY_GOOD, 60, GC_FRANCE,     "F",   "France"      },
    { 34, 0, VERY_GOOD, 60, GC_ITALY,      "I",   "Italy"       },
    { 35, 0, MEDIUM,     1, GC_MALTA,      "M",   "Malta"       },
    { 36, 0, GOOD,      10, GC_PORTUGAL,   "P",   "Portugal"    },
    { 37, 0, VERY_GOOD, 40, GC_SPAIN,      "E",   "Spain"       },

    { 38, 0, VERY_GOOD, 10, GC_BELGIUM,    "B",   "Belgium"     },
    { 39, 0, VERY_GOOD,  1, GC_LUXEMBOURG, "LUX", "Luxembourg"  },
    { 40, 0, VERY_GOOD, 14, GC_NETHERLANDS,"NL",  "the Netherlands" },
    { 41, 0, GOOD,       1, GC_EAST_FRISIA,"FRI", "East Frisia" },
    { 42, 0, VERY_GOOD, 80, GC_GERMANY,    "D",   "Germany"     },
    { 43, 0, VERY_GOOD,  8, GC_AUSTRIA,    "A",   "Austria"     },
    { 44, 0, VERY_GOOD,  7, GC_SWISS,      "CH",  "Swiss"       },

    { 45, 0, VERY_GOOD,  1, GC_ICELAND,    "ICE", "Iceland"     },
    { 46, 0, VERY_GOOD,  5, GC_DENMARK,    "DK",  "Denmark"     },
    { 47, 0, GOOD,       4, GC_NORWAY,     "N",   "Norway"      },
    { 48, 0, VERY_GOOD,  8, GC_SWEDEN,     "S",   "Sweden"      },
    { 49, 0, GOOD,       5, GC_FINLAND,    "FIN", "Finland"     },
    { 50, 0, GOOD,       2, GC_ESTONIA,    "EST", "Estonia"     },
    { 51, 0, GOOD,       2, GC_LATVIA,     "LTV", "Latvia"      },
    { 52, 0, GOOD,       3, GC_LITHUANIA,  "LIT", "Lithuania"   },

    { 53, 0, GOOD,      35, GC_POLAND,     "PL",  "Poland"      },
    { 54, 0, GOOD,       8, GC_CZECH_REP,  "CZ",  "Czech Republic" },
    { 55, 0, GOOD,       7, GC_SLOVAKIA,   "SK",  "Slovakia"    },
    { 56, 0, GOOD,      11, GC_HUNGARY,    "H",   "Hungary"     },
    { 57, 0, VERY_GOOD, 22, GC_ROMANIA,    "RO",  "Romania"     },
    { 58, 0, GOOD,       9, GC_BULGARIA,   "BG",  "Bulgaria"    },

    { 59, 0, MEDIUM,     4, GC_BOSNIA,     "BIH","Bosnia and Herzegovina" },
    { 60, 0, GOOD,       5, GC_CROATIA,    "CRO", "Croatia"     },
    { 61, 0, MEDIUM,     1, GC_KOSOVO,     "KOS", "Kosovo"      },
    { 62, 0, MEDIUM,     2, GC_MACEDONIA,  "MK",  "Macedonia"   },
    { 63, 0, MEDIUM,     1, GC_MONTENEGRO, "MON", "Montenegro"  },
    { 64, 0, MEDIUM,     9, GC_SERBIA,     "SER", "Serbia"      },
    { 65, 0, MEDIUM,     2, GC_SLOVENIA,   "SLO", "Slovenia"    },
    { 66, 0, GOOD,       3, GC_ALBANIA,    "AL",  "Albania"     },
    { 67, 0, GOOD,      10, GC_GREECE,     "GR",  "Greece"      },

    { 68, 0, GOOD,     100, GC_RUSSIA,     "RUS", "Russia"      },
    { 69, 0, MEDIUM,    10, GC_BELARUS,    "BY",  "Belarus"     },
    { 70, 0, MEDIUM,     4, GC_MOLDOVA,    "MOL", "Moldova"     },
    { 71, 0, MEDIUM,    45, GC_UKRAINE,    "UKR", "Ukraine"     },
    { 72, 0, MEDIUM,     3, GC_ARMENIA,    "ARM", "Armenia"     },
    { 73, 0, MEDIUM,     4, GC_AZERBAIJAN, "AZE", "Azerbaijan"  },
    { 74, 0, MEDIUM,     5, GC_GEORGIA,    "GEO", "Georgia"     },
    { 75, 0, MEDIUM,    15,GC_KAZAKH_UZBEK,"KAZ", "Kazakhstan/Uzbekistan" },

    { 76, 0, GOOD,      55, GC_TURKEY,     "TUR", "Turkey"      },
    { 77, 0, GOOD,      80, GC_ARABIA,     "AR",  "Arabia/Persia" },
    { 78, 0, MEDIUM,     4, GC_ISRAEL,     "ISR", "Israel"      },
    { 79, 0, VERY_GOOD,300, GC_CHINA,      "CHN", "China"       },
    { 80, 0, GOOD,     250, GC_INDIA,      "IND", "India/Sri Lanka" },
    { 81, 0, GOOD,      35, GC_JAPAN,      "JAP", "Japan"       },
    { 82, 0, GOOD,      12, GC_KOREA,      "KOR", "Korea"       },
    { 83, 0, GOOD,      17, GC_VIETNAM,    "VN",  "Vietnam"     },
    { 84, 0, MEDIUM,     1,    -1L,        "XX",  "other countries" },
    {  0, 0,  0,         0,    -1L,        NULL,    NULL        }
  };


static char *umlauts_unicode[]
  =  {
       "256",  "<A/>",
       "257",  "<a/>",
       "258",  "<A^>",
       "258",  "<>",   /** <A^> **/
       "259",  "<a^>",
       "259",  "<>",   /** <a^> **/
       "260",  "<A,>",
       "261",  "<a,>",
       "262",  "<C>",
       "263",  "<c>",
       "268",  "<C^>",
       "268",  "<CH>",
       "269",  "<c^>",
       "269",  "<ch>",
       "271",  "<d>",
       "272",  "<>",
       "272",  "<DJ>",
       "273",  "<>",
       "273",  "<dj>",
       "274",  "<E/>",
       "275",  "<e/>",
       "278",  "<E>",
       "279",  "<e>",
       "280",  "<E,>",
       "281",  "<e,>",
       "282",  "<E^>",
       "282",  "<>",
       "283",  "<e^>",
       "283",  "<>",
       "286",  "<G^>",
       "287",  "<g^>",
       "290",  "<G,>",
       "291",  "<g>",
       "298",  "<I/>",
       "299",  "<i/>",
       "304",  "<I>",
       "305",  "<i>",
     /***
       "306",  "<IJ>",
       "307",  "<ij>",
     ***/
       "310",  "<K,>",
       "311",  "<k,>",
       "315",  "<L,>",
       "316",  "<l,>",
       "317",  "<L>",
       "318",  "<l>",
       "321",  "<L/>",
       "322",  "<l/>",
       "325",  "<N,>",
       "326",  "<n,>",
       "327",  "<N^>",
       "328",  "<n^>",
       "336",  "<>",
       "337",  "<>",
       "338",  "<OE>",
       "338",  "",   /** <OE> **/
       "339",  "<oe>",
       "339",  "",   /** <oe> **/
       "344",  "<R^>",
       "345",  "<r^>",
       "350",  "<S,>",
       "351",  "<s,>",
       "352",  "<S^>",
       "352",  "",   /** <S^> **/
       "352",  "<SCH>",
       "352",  "<SH>",
       "353",  "<s^>",
       "353",  "",   /** <s^> **/
       "353",  "<sch>",
       "353",  "<sh>",
       "354",  "<T,>",
       "355",  "<t,>",
       "357",  "<t>",
       "362",  "<U/>",
       "363",  "<u/>",
       "366",  "<U>",
       "367",  "<u>",
       "370",  "<U,>",
       "371",  "<u,>",
       "379",  "<Z>",
       "380",  "<z>",
       "381",  "<Z^>",
       "382",  "<z^>",
       NULL,    NULL
     };


int initialize_gender (TSRMLS_D)
{
  int  i,k;
  char *s,*s2,*s3;

  if (! (GENDER_G(internal_mode) & GENDER_INITIALIZED))
    {
      if ((int)strlen (letters_a_to_z) > 26)
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: %s  is not allowed\n",
                  "strlen (letters_a_to_z) > 26");
            }
          return (-1);
        }

      if ((int)strlen (letters_a_to_z) != (int)strlen (letters_A_to_Z))
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: %s  is not allowed\n",
                  "strlen(letters_a_to_z) != strlen(letters_a_to_z)");
            }
          return (-1);
        }
      if ((int)strlen (umlaut_lower) != (int)strlen (umlaut_upper))
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: %s  is not allowed\n",
                  "strlen(umlaut_lower) != strlen(umlaut_upper)");
            }
          return (-1);
        }

      if ((int)strlen (umlaut_lower) != (int)strlen (umlaut_sort)
      ||  (int)strlen (umlaut_lower) != (int)strlen (umlaut_sort2))
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: %s  is not allowed\n",
                  "strlen(umlaut_sort*) != strlen(umlaut_lower)");
            }
          return (-1);
        }

      GENDER_G(internal_mode) |= GENDER_INITIALIZED;

      /****  generate arrays "GENDER_G(sortchar)", "GENDER_G(sortchar2)" and "GENDER_G(upperchar)"  ****/
      for (i=0; i< HASH_COUNT; i++)
        {
          GENDER_G(sortchar)[i] = (char) i;
          GENDER_G(sortchar2)[i] = '\0';
          GENDER_G(upperchar)[i] = (char) i;
        }
      GENDER_G(sortchar) [(unsigned char) '-'] = ' ';
      GENDER_G(sortchar) [(unsigned char) '\''] = '';

      s = CHARS_TO_IGNORE;
      for (i=0; s[i] != '\0'; i++)
        {
          k = (unsigned char) s[i];
          GENDER_G(sortchar)[k] = IS_CHAR_TO_IGNORE;
        }

      s = letters_a_to_z;
      s2 = letters_A_to_Z;

      for (i=0; s[i] != '\0'; i++)
        {
          k = (unsigned char) s2[i];
          GENDER_G(sortchar)[k] = s2[i];
          GENDER_G(upperchar)[k] = s2[i];

          k = (unsigned char) s[i];
          GENDER_G(sortchar)[k] = s2[i];
          GENDER_G(upperchar)[k] = s2[i];
        }

      s = umlaut_lower;
      s2 = umlaut_upper;

      for (i=0; s[i] != '\0'; i++)
        {
          k = (unsigned char) s2[i];  /** "s2" **/
          GENDER_G(upperchar)[k] = s2[i];

          k = (unsigned char) s[i];   /** "s" **/
          GENDER_G(upperchar)[k] = s2[i];
        }

      s = umlaut_lower;
      s2 = umlaut_sort;
      s3 = umlaut_sort2;

      for (i=0; s[i] != '\0'; i++)
        {
          k = (unsigned char) s[i];
          GENDER_G(sortchar)[k] = s2[i];
          if (s3[i] != ' ')
            {
              GENDER_G(sortchar2)[k] = s3[i];
            }

          k = (unsigned char) umlaut_upper[i];
          GENDER_G(sortchar)[k] = s2[i];
          if (s3[i] != ' ')
            {
              GENDER_G(sortchar2)[k] = s3[i];
            }
        }
    }

  return (0);
}


static int copycut (char dest[], char src[], int len TSRMLS_DC)
{
  int i,k,n;
  char c;

  k = 0;
  n = 0;
  i = 1;
  while ((c = src[k]) != '\0'  &&  c != '\n'
  &&  (k < len-1  ||  len < 0))
    {
      if (c == ' ')
        {
          i--;
        }
      else
        {
          i = 2;
        }

      if (i > 0)
        {
          dest[n] = c;
          n++;
        }
      k++;
    }

  if (n > 0  &&  dest[n-1] == ' ')
    {
      /****  delete trailing space  ****/
      n--;
    }
  dest[n] = '\0';

  return (n);
}



static int strcmp_search (char text[], char text_2[],
     int compare_mode, char umlaut_info, int *diff TSRMLS_DC)

/****  Internal function for "string compare":          ****/
/****  Compare strings case-insensitive and ignore      ****/
/****  different spellings of umlauts.                  ****/
/****  Return value is the same as for normal "strcmp". ****/
/****  First arg  = "external" name                     ****/
/****  Second arg = "internal" name (with "unicode" chars) ****/

/****  Important bit masks for "compare_mode":          ****/
/****  EXPAND_PARAM_1  :   expand umlauts in "text"     ****/
/****  EXPAND_PARAM_2  :   expand umlauts in "text_2"   ****/
/****  COMPARE_ABBREVIATION :  compare abbreviation     ****/
/****                          (e.g. "C." and "Carl")   ****/

/****  This function converts "src" to upper case,      ****/
/****  compresses successive blanks and (if necessary)  ****/
/****  expands umlauts, i.e. "" is converted to "AE".  ****/
/****  The conversion is done "on demand" using the     ****/
/****  arrays "GENDER_G(sortchar)" and "GENDER_G(sortchar2)". The second    ****/
/****  array is needed only for umlauts and contains    ****/
/****  null values for all other chars.                 ****/
{
  int  i,i2,k,n;
  char c=0,c2=0;

  if (umlaut_info == COMPRESSED_UMLAUTS_FOR_SORTING)
    {
      compare_mode &= ~EXPAND_PARAM_2;
    }
  if (umlaut_info == EXPANDED_UMLAUTS_FOR_SORTING)
    {
      compare_mode |= EXPAND_PARAM_2;
    }

  i = 0;
  i2 = 0;
  k = 0;
  while (GENDER_G(sortchar) [(unsigned char) *text_2] == IS_CHAR_TO_IGNORE)
    {
      text_2++;
    }

  while (*text != '\0'  &&  *text_2 != '\0')
    {
      /****  find next char in first string             ****/
      /****  (without "while (c == IS_CHAR_TO_IGNORE) ...") ****/
      c = '\0';
      if (i != 0)
        {
          if (compare_mode & EXPAND_PARAM_1)
            {
              c = GENDER_G(sortchar2) [(unsigned char) *text];
            }
          i = 0;
          text++;
        }

      if (c == '\0')
        {
          c = GENDER_G(sortchar) [(unsigned char) *text];
          i = 1;

          if ((c == ' '  ||  c == '-')
          &&  (compare_mode & IGNORE_SEPARATOR))
            {
              text++;
              c = GENDER_G(sortchar) [(unsigned char) *text];
            }
        }

      /****  find next char in second string  ****/

      c2 = '\0';
      if (i2 != 0)
        {
          if (compare_mode & EXPAND_PARAM_2)
            {
              c2 = GENDER_G(sortchar2) [(unsigned char) *text_2];
            }
          i2 = 0;
          text_2++;
          while (GENDER_G(sortchar) [(unsigned char) *text_2] == IS_CHAR_TO_IGNORE)
            {
              text_2++;
            }
        }

      if (c2 == '\0')
        {
          c2 = GENDER_G(sortchar) [(unsigned char) *text_2];
          i2 = 1;

          if ((c2 == ' '  ||  c2 == '-'  ||  *text_2 == '+')
          &&  (compare_mode & IGNORE_SEPARATOR))
            {
              text_2++;
              while (GENDER_G(sortchar) [(unsigned char) *text_2] == IS_CHAR_TO_IGNORE)
                {
                  text_2++;
                }
              c2 = GENDER_G(sortchar) [(unsigned char) *text_2];
            }
        }

      if (c2 == '+')
        {
          /****  "internal separator" ('+') found  ****/
          c2 = '-';
          if (c == ' '  ||  c == '-')
            {
              c2 = c;
              continue;  /****  important  ****/
            }
          else if (i == 1  &&  i2 == 1
          &&  GENDER_G(sortchar2) [(unsigned char) *text] == '\0')
            {
              text_2++;
              while (GENDER_G(sortchar) [(unsigned char) *text_2] == IS_CHAR_TO_IGNORE)
                {
                  text_2++;
                }
              c2 = GENDER_G(sortchar) [(unsigned char) *text_2];
            }
        }

      /****  compare chars  ****/
      if ((compare_mode & COMPARE_ABBREVIATION)
      &&  c == '.'  &&  c2 != '\0')
        {
          /****  strings match like (e.g.) "St." and "Steven"  ****/
          c = c2;
          break;
        }
      if (c != c2)
        {
          /****  strings don't match  ****/
          break;
        }

      if (diff != NULL  &&  i == 1  &&  i2 == 1)
        {
          if (GENDER_G(upperchar) [(unsigned char) *text] != GENDER_G(upperchar) [(unsigned char) *text_2])
            {
              if (compare_mode & GENDER_COMPARE_EXACT)
                {
                  /****  strings don't match exactly  ****/
                  k |= 1024;
                }

              c = GENDER_G(sortchar2) [(unsigned char) *text];
              c2 = GENDER_G(sortchar2) [(unsigned char) *text_2];

              if (c != c2)
                {
                  k++;
                  if (c == '\0')
                    {
                      k |= 128;
                    }
                  if (c2 == '\0')
                    {
                      k |= 256;
                    }
                  if (c != '\0'  &&  c2 != '\0')
                    {
                      k |= 1024;
                    }
                }

              if (c == '\0')
                {
                  c = GENDER_G(sortchar) [(unsigned char) *(text+1)];

                  if ((c == ' '  ||  c == '-')
                  &&  (compare_mode & IGNORE_SEPARATOR))
                    {
                      c = GENDER_G(sortchar) [(unsigned char) *(text+2)];
                    }
                }

              if (c2 == '\0')
                {
                  n = 1;
                  while (GENDER_G(sortchar) [(unsigned char) *(text_2+n)] == IS_CHAR_TO_IGNORE)
                    {
                      n++;
                    }
                  c2 = GENDER_G(sortchar) [(unsigned char) *(text_2+n)];
                }

              if (c != c2
              && ((compare_mode & EXPAND_PARAM_2)
	      ||  ! (compare_mode & GENDER_ALLOW_COMPRESSED_UMLAUTS)))
                {
                  /****  strings don't match (e.g. "Kte" vs "Kte")  ****/
                  k |= 1024;
                }
            }
        }
    }

  if (diff != NULL)
    {
      if (c != c2  ||  (k & (128+256)) == 128+256)
        {
          /****  strings don't match (e.g. "Kat" vs "Kte")  ****/
          k |= 1024;
        }
      *diff = k;
    }

  if (*text == '\0'  ||  *text_2 == '\0')
    {
      if (c == c2)
        {
          if (i != 0  &&  *text != '\0'
          && (GENDER_G(sortchar2) [(unsigned char) *text] == '\0'
          ||  ! (compare_mode & EXPAND_PARAM_1)))
            {
              text++;
            }
          if (i2 != 0  &&  *text_2 != '\0'
          && (GENDER_G(sortchar2) [(unsigned char) *text_2] == '\0'
          ||  ! (compare_mode & EXPAND_PARAM_2)))
            {
              text_2++;
            }
        }

      return ((unsigned char) *text - (unsigned char) *text_2);
    }

  return ((unsigned char) c - (unsigned char) c2);
}

static void read_line (php_stream *fr, char text[], int len TSRMLS_DC)
{
  text[0] = '\0';
  php_stream_gets(fr, text, len-2);
  text[len-1] = '\0';
}



static long binary_search (php_stream *fr, char search_name[],
    int offset, int len_s, int compare_mode TSRMLS_DC)

/****  Do a binary search for the fist match             ****/
/****  (all lines in data file must have equal length)   ****/
/****  Return value == file pos of first match           ****/
/****                  ( < 0L: no match found)           ****/
/****  "compare_mode" is the same as for "strcmp_search" ****/
{
  long p,p1,p2;
  int  i,n, returned_len;
  char text [MAX_LINE_SIZE+1];
  char temp [MAX_LINE_SIZE+1];

  if (GENDER_G(line_size) == 0  ||  GENDER_G(record_count) == 0L)
    {
      /****  determine length of first line  ****/
      i = php_stream_seek(fr, 0L, SEEK_SET);
      if (i != 0)
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: Could not position in data file\n");
            }
          return (-10L);
        }

      read_line(fr, text, MAX_LINE_SIZE+1 TSRMLS_CC);
      if (strncmp (text, CHECK_STRING, (int)strlen(CHECK_STRING)) != 0)
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              i = (int)strlen (text);
              while (i > 0  &&  (text[i-1] == '\n'  ||  text[i-1] == '\r'))
                {
                  i--;
                }
              text[i] = '\0';

              php_printf ("Error: Invalid version of data file.\n");
              php_printf ("File header is:  \"%s\"\n", text);
              php_printf ("(this should be:  \"%s\").\n", CHECK_STRING);
            }
          return (-10L);
        }

      /****  get line size  ****/
      GENDER_G(line_size) = (int) php_stream_tell(fr);

      /****  determine file size  ****/
      i = php_stream_seek(fr,0L, SEEK_END);
      if (i != 0)
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: Could not position in data file\n");
            }
          return (-10L);
        }

      p = php_stream_tell(fr);
      GENDER_G(record_count) = (long) ((p+1L) / GENDER_G(line_size));

      /****  reset cache variables  ****/
      for (i=0; i< CACHE_SIZE; i++)
        {
          GENDER_G(cache_p)[i] = -1L;
          strcpy (GENDER_G(cache_text)[i],"");
        }
    }

  /****  start binary search  ****/
  p1 = 0L;
  p2 = GENDER_G(record_count);
  i = -1;
  n = -1;

  while (p1 <= p2)
    {
      p = (long) ((p1+p2) / 2L);
      n++;

      if (n < CACHE_SIZE  &&  p == GENDER_G(cache_p)[n])
        {
          /****  read data from cache  ****/
          strcpy (text, GENDER_G(cache_text)[n]);
        }
      else
        {
          /****  read from data file  ****/
          i = php_stream_seek(fr, (GENDER_G(line_size) * p), SEEK_SET);
          if (i != 0)
            {
              if (GENDER_G(internal_mode) & TRACE_GENDER)
                {
                  php_printf ("Could not position in data file\n");
                }
              return (-1L);
            }

          read_line(fr, text, MAX_LINE_SIZE+1 TSRMLS_CC);

          if (n < CACHE_SIZE)
            {
              /****  copy data into cache  ****/
              GENDER_G(cache_p)[n] = p;
              strcpy (GENDER_G(cache_text)[n], text);
            }
        }

      strcpy (temp,"");
      if (text[0] != '#'  &&  (int)strlen (text) > offset)
        {
          copycut (temp, text+offset, len_s TSRMLS_CC);
        }

      if (GENDER_G(internal_mode) & TRACE_GENDER)
        {
          php_printf ("Range = line %ld - %ld,  guess = %ld ('%s')\n",
              p1+1L, p2+1L, p+1L, temp);
        }

      /****  compare first names  ****/
      i = strcmp_search (search_name, temp, (compare_mode | IGNORE_SEPARATOR),
              text[POS_UMLAUT_INFO], NULL TSRMLS_CC);

      if (i == 0)
        {
          if (p == p1)
            {
              /****  first match has been found  ****/
              break;
            }
          p2 = p;
        }
      else if (i < 0)
        {
          p2 = p - 1L;
        }
      else
        {
          p1 = p + 1L;
          p++;   /**** important  ****/
        }
    }

  if (i == 0)
    {
      /****  match has been found  ****/
      if (GENDER_G(internal_mode) & TRACE_GENDER)
        {
          php_printf ("Result: name '%s' found\n", temp);
        }
      return (GENDER_G(line_size) * p);
    }

  /****  no match found  ****/
  if (GENDER_G(internal_mode) & TRACE_GENDER)
    {
      php_printf ("Result: name '%s' not found\n", search_name);
    }
  return (-1L);
}



static int conv_internal_result (char *text, int compare_mode)
{
  if (compare_mode & SEARCH_GENDER)
    {
      if (compare_mode & SEARCH_FIRST)
        {
          if (strncmp (text,"1F",2) == 0)  return (IS_FEMALE);
          if (strncmp (text,"1M",2) == 0)  return (IS_MALE);
        }

      if (strncmp (text,"F ",2) == 0)  return (IS_FEMALE);
      if (strncmp (text,"1F",2) == 0)  return (IS_MOSTLY_FEMALE);
      if (strncmp (text,"?F",2) == 0)  return (IS_MOSTLY_FEMALE);

      if (strncmp (text,"M ",2) == 0)  return (IS_MALE);
      if (strncmp (text,"1M",2) == 0)  return (IS_MOSTLY_MALE);
      if (strncmp (text,"?M",2) == 0)  return (IS_MOSTLY_MALE);

      if (strncmp (text,"? ",2) == 0)  return (IS_UNISEX_NAME);

      return (NAME_NOT_FOUND);
    }

  if (compare_mode & SEARCH_EQUIV)
    {
      if (strncmp (text,"= ",2) == 0)  return (EQUIVALENT_NAMES);

      return (NAME_NOT_FOUND);
    }

  return (INTERNAL_ERROR_GENDER);
}


static int get_frequency (char c)
{
  int x;

  x = 0;
  if (isdigit ((unsigned char) c))
    {
      x = (unsigned char) c - '0';
    }
  else if (c != ' ')
    {
      x = (unsigned char) c - 'A' + 10;
      if (x > 13  ||  x < 0)
        {
          x = 13;
        }
    }

  return (x);
}


static int multiply_frequency (int a, int b)
{
  int x;

  if (a > 13)
    {
      a = 13;
    }
  if (b > 13)
    {
      b = 13;
    }
  
  if (a < 0)
    {
      return (b);
    }
  if (b < 0)
    {
      return (a);
    }

  x = a + b - 15;
  if (x < 0)
    {
      x = 0;
    }
  if (x == 0  &&  (a*b > 0  ||  a+b >= 9))
    {
      x = 1;
    }

  return (x);
}



static void trace_info
     (char text[], char name[], char name2[], int res, char data[] TSRMLS_DC)
/****  output trace info  ****/
{
  int  i,x;
  int  pos_0;
  int  freq,freq_2;
  long p,count;
  char *s,*s2;

  switch (res)
    {
      case IS_FEMALE        :  s = "is female";            break;
      case IS_MOSTLY_FEMALE :  s = "is mostly female";     break;
      case IS_MALE          :  s = "is male";              break;
      case IS_MOSTLY_MALE   :  s = "is mostly male";       break;
      case IS_A_COUPLE      :  s = "is a couple";          break;
      case IS_UNISEX_NAME   :  s = "is unisex name";       break;

      case EQUIVALENT_NAMES :  s = "names are equivalent"; break;
      case NOT_EQUAL_NAMES  :  s = "names are not equal";  break;
      case NAME_NOT_FOUND   :  s = "name not found";       break;
      case ERROR_IN_NAME    :  s = "error in name";        break;

      case INTERNAL_ERROR_GENDER :  s = "internal error";  break;
      default                    :  s = "unknown error";   break;
    }

  if (name2 != NULL)
    {
      php_printf ("%s '%s'%s':  '%s'", text, name,name2, s);
    }
  else if (text[0] != '\0')
    {
      php_printf ("%s '%s':  '%s'", text, name, s);
    }
  else
    {
      php_printf ("%s:  '%s'", name, s);
    }

  if (data != NULL  ||  (GENDER_G(internal_mode) & GENDER_GET_COUNTRY))
    {
      /****  print country  ****/
      freq = 0;
      freq_2 = 0;
      pos_0 = -1;
      count = 1L;

      for (i=0; gc_data[i].country_text != NULL; i++)
        {
          x = 0;
          if (data != NULL  &&  *data != '\0')
            {
              x = get_frequency (data[gc_data[i].pos]);
            }
          else if ((GENDER_G(internal_mode) & GENDER_GET_COUNTRY)
          &&  gc_data[i].n > 0)
            {
              x = (int) (gc_data[i].n % 16);
            }

          if (x > freq)
            {
              freq = x;
            }
        }

      freq_2 = (int)((freq+1) / 2);
      if (freq_2 <= 1)
        {
          freq_2 = freq;
        }
      if (freq_2 < freq - 3)
        {
          freq_2 = freq - 3;
        }
      freq = 0;

      s = " (country =";
      for (i=0; gc_data[i].country_text != NULL; i++)
        {
          x = 0;

          if (data != NULL  &&  *data != '\0')
            {
              x = get_frequency (data[gc_data[i].pos]);
            }
          else if ((GENDER_G(internal_mode) & GENDER_GET_COUNTRY)
          &&  gc_data[i].n > 0)
            {
              x = (int) (gc_data[i].n % 16);
            }

          if (x > 0)
            {
              if (! (GENDER_G(internal_mode) & GENDER_TRACE_ONE_COUNTRY_ONLY))
                {
                  if (GENDER_G(internal_mode) & GENDER_TRACE_SHORT_COUNTRY_NAME)
                    {
                      s2 = gc_data[i].country_short;
                    }
                  else
                    {
                      s2 = gc_data[i].country_text;
                    }
                  php_printf (" %s %s[%d]", s,s2,x);
                  s = "or";
                }

              if (x >= freq_2)
                {
                  p = (long) gc_data[i].weight;
                  if (p < 1L)
                    {
                      p = 1L;
                    }
                  if (gc_data[i].n & 1024)
                    {
                      /****  country has been found  ****/
                      p = 5L * p;
                    }
                  if (x > 1)
                    {
                      p *= (1L << x);
                    }

                  if (p > count  ||  pos_0 < 0
                  || (p == count  &&  x > freq))
                    {
                      pos_0 = i;
                      freq = x;
                      count = p;
                    }
                }
            }
        }

      if (! (GENDER_G(internal_mode) & GENDER_TRACE_ONE_COUNTRY_ONLY))
        {
          if (strcmp (s,"or") == 0)
            {
              php_printf (")");
            }
        }
      else
        {
          if (pos_0 >= 0)
            {
              if (GENDER_G(internal_mode) & GENDER_TRACE_SHORT_COUNTRY_NAME)
                {
                  s2 = gc_data[pos_0].country_short;
                }
              else
                {
                  s2 = gc_data[pos_0].country_text;
                }
              php_printf (" (country=%s)", s2);
            }
        }
    }

  php_printf ("\n");
}


static int internal_search
     (char search_name[], int compare_mode, int country TSRMLS_DC)
/****  This function searches a single given name  ****/
{
  int  i,n,x;
  int  k,k0;
  int  freq,freq_2;
  int  pos_0,pos_x[7];
  int  freq_x[7];
  int  len,res;
  long count,count_2,count_x[7];
  long p,pos_f;
  char text [MAX_LINE_SIZE+1];
  char temp [DATA_NAME_LENGTH+1];

  if (GENDER_G(internal_mode) & TRACE_GENDER)
    {
      php_printf ("Searching for name '%s'", search_name);

      /****  check for country  ****/
      for (i=0; gc_data[i].country_text != NULL; i++)
        {
          if (country == gc_data[i].gc_country)
            {
              php_printf ("  (country = %s)", gc_data[i].country_text);
            }
        }
      php_printf ("\n");
    }

  if (GENDER_G(f_names) == NULL)
    {
      if(gender_connect_to_source(GENDER_G(dsn) TSRMLS_CC) == INTERNAL_ERROR_GENDER)
        {
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("Error: could not open data file '%s'\n", GENDER_G(dsn));
            }
          return (INTERNAL_ERROR_GENDER);
        }
    }

  if (compare_mode & SEARCH_GENDER)
    {
     i = (int)strlen (search_name);
     if (i == 0)
       {
         return (IS_UNISEX_NAME);
       }

     if (i >= 2  &&  search_name[i-1] == '.')
       {
         compare_mode |= COMPARE_ABBREVIATION;
         GENDER_G(internal_mode) &= ~GENDER_GET_COUNTRY;
       }
    }

  if (compare_mode & UMLAUT_EXPAND)
    {
      compare_mode |= EXPAND_PARAM_1;
    }

  pos_f = binary_search (GENDER_G(f_names), search_name, DATA_NAME_POS,
            DATA_NAME_LENGTH+1, compare_mode TSRMLS_CC);

  if (pos_f < 0L)
    {
      if (pos_f == -10L)
        {
          return (INTERNAL_ERROR_GENDER);
        }
      return (NAME_NOT_FOUND);
    }

  /****  read one or more first names in data file  ****/
  res = NAME_NOT_FOUND;
  php_stream_seek(GENDER_G(f_names), pos_f, SEEK_SET);

  for (x=0; x<7; x++)
    {
      pos_x[x] = 0;
      freq_x[x] = 0;
      count_x[x] = 0L;
    }
  k0 = 500;
  i = 0;

  do
    {
      read_line (GENDER_G(f_names),text, MAX_LINE_SIZE+1 TSRMLS_CC);
      len = (int)strlen (text);

      strcpy (temp,"");
      if (text[0] != '#'  &&  len > DATA_NAME_POS)
        {
          copycut (temp, text+DATA_NAME_POS, DATA_NAME_LENGTH+1 TSRMLS_CC);
        }

      /****  compare first names  ****/
      n = strcmp_search (search_name, temp, compare_mode,
              text [POS_UMLAUT_INFO], &k TSRMLS_CC);

      if (n == 0       /****  names are matching     and     ****/
      &&  k <= k0)     /****  umlauts don't differ too much  ****/
        {
          if (k < k0)
            {
              k0 = k;
              i = 0;
            }

          res = conv_internal_result (text, compare_mode);
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              trace_info ("evaluating name", temp,NULL, res,text TSRMLS_CC);
            }

          if (compare_mode & COMPARE_ABBREVIATION)
            {
              if (res == IS_FEMALE)  res = IS_MOSTLY_FEMALE;
              if (res == IS_MALE)  res = IS_MOSTLY_MALE;

              if ((int)strlen (search_name) <= 3
              ||  (i & 16)  ||  ((i & 2)  &&  (i & 8)))
                {
                  return (IS_UNISEX_NAME);
                }
            }

          /****  check country  ****/
          freq = 0;
          freq_2 = 0;
          pos_0 = -1;
          count = 1L;

          for (n=0; gc_data[n].country_text != NULL; n++)
            {
              if (len <= (int) gc_data[n].pos)
                {
                  /****  line is too short  ****/
                  return (INTERNAL_ERROR_GENDER);
                }

              if (text [gc_data[n].pos] != ' ')
                {
                  x = get_frequency (text[gc_data[n].pos]);

                  if (x > freq)
                    {
                      freq = x;
                    }
                }
            }

          freq_2 = (int)((freq+1) / 2);
          if (freq_2 <= 1)
            {
              freq_2 = freq;
            }
          if (freq_2 < freq - 3)
            {
              freq_2 = freq - 3;
            }

          freq = 0;
          for (n=0; gc_data[n].country_text != NULL; n++)
            {
              if (text [gc_data[n].pos] != ' ')
                {
                  x = get_frequency (text[gc_data[n].pos]);

                  if (x >= freq_2)    /***  freq_2 !!  ****/
                    {
                      p = (long) gc_data[i].weight;
                      if (p < 1L)
                        {
                          p = 1L;
                        }
                      if (gc_data[i].n & 1024)
                        {
                          /****  country has been found  ****/
                          p = 5L * p;
                        }
                      if (x > 1)
                        {
                          p *= (1L << x);
                        }

                      if (p > count  ||  pos_0 < 0
                      || (p == count  &&  x > freq))
                        {
                          pos_0 = i;
                          freq = x;
                          count = p;
                        }
                    }

                  if (x > (int) (gc_data[n].n % 16))
                    {
                      x -= (int) (gc_data[n].n % 16);
                      gc_data[n].n += x;
                    }
                }
            }

          switch (res)
            {
              case IS_FEMALE        :  n =  1;  x = 0;  break;
              case IS_MOSTLY_FEMALE :  n =  2;  x = 1;  break;
              case IS_MALE          :  n =  4;  x = 2;  break;
              case IS_MOSTLY_MALE   :  n =  8;  x = 3;  break;
              case IS_UNISEX_NAME   :  n = 16;  x = 4;  break;
              case EQUIVALENT_NAMES :  n = 32;  x = 5;  break;
              default               :  n = 64;  x = 6;  break;
            }

          i |= n;
          if (count > count_x[x])
            {
              pos_x[x] = pos_0;
              freq_x[x] = freq;
              count_x[x] = count;
            }
        }

      /****  "strcmp_search" with mode "IGNORE_SEPARATOR" (!!)  ****/
      n = strcmp_search (search_name, temp, (compare_mode | IGNORE_SEPARATOR),
             text [POS_UMLAUT_INFO], & k TSRMLS_CC);
    }while (n == 0);


  n = 0;
  count = 0L;
  count_2 = 0L;
  freq = 0;
  freq_2 = 0;

  for (x=0; x<7; x++)
    {
      if (count_x[x] > count
      || (count_x[x] == count  &&  freq_x[x] > freq))
        {
          n = 1 << x;
          count_2 = count;
          count = count_x[x];
          freq_2 = freq;
          freq = freq_x[x];
        }
      else if (count_x[x] > count_2
      || (count_x[x] == count_2  &&  freq_x[x] > freq_2))
        {
          count_2 = count_x[x];
          freq_2 = freq_x[x];
        }
    }

  if (freq_2 > 0)
    {
      /****  calculate the (truncated) binary logarithm of (count/count_2)  ****/
      x = 5;
      if (count > 0L  &&  count < count_2)
        {
          x = 0;
          while (count < count_2)
            {
              count *= 2L;
              x--;
            }
        }
      else if (count_2 > 0  &&  count_2 <= count)
        {
          x = -1;
          while (count_2 <= count)
            {
              count_2 *= 2;
              x++;
            }
        }
      k = freq - freq_2;

      if (country != GC_ANY_COUNTRY
      &&  k+k + x >= 10)
        {
          /****  ignore other countries  ****/
          i = n;
        }
      else if (k+k + x >= 6)
        {
          if (n == 1  ||  n == 2)
            {
              i &= ~(1|4|8|16);
              i |= 2;
            }
          else if (n == 4  ||  n == 8)
            {
              i &= ~(1|2|4|16);
              i |= 8;
            }
        }
    }

  if ((i & (1|2|16))  &&  (i & (4|8|16)))
    {
      return (IS_UNISEX_NAME);
    }

  if (i & 2)
    {
      return (IS_MOSTLY_FEMALE);
    }
  if (i & 8)
    {
      return (IS_MOSTLY_MALE);
    }

  if (i == 1)
    {
      return (IS_FEMALE);
    }
  if (i == 4)
    {
      return (IS_MALE);
    }

  return (res);
}





/************************************************************/
/****  "external" functions  ********************************/
/************************************************************/


int get_gender (char first_name[], int compare_mode, int country TSRMLS_DC)
/****  This function determines the gender of a given first name  ****/

/****  Result: IS_FEMALE        :  female first name            ****/
/****          IS_MOSTLY_FEMALE :  mostly female first name     ****/
/****          IS_MALE          :  male first name              ****/
/****          IS_MOSTLY_MALE   :  mostly male first name       ****/
/****          IS_A_COUPLE      :  name is both male and female ****/
/****          IS_UNISEX_NAME   :  unisex name (can be any gender) ****/
/****          NAME_NOT_FOUND   :  name not found               ****/
/****          ERROR_IN_NAME    :  name contains an error       ****/
{
  int  i,k,n,x;
  int  gender,res;
  char c;
  char temp [LENGTH_FIRST_NAME+52];

  if (first_name[0] == '\0')
    {
      return (IS_UNISEX_NAME);
    }

  compare_mode &= ~(SEARCH_EQUIV | SEARCH_FIRST);
  compare_mode |= SEARCH_GENDER;

  copycut (temp, first_name, LENGTH_FIRST_NAME+51 TSRMLS_CC);

  /****  reset statistics  ****/
  for (n=0; gc_data[n].country_text != NULL; n++)
    {
      gc_data[n].n = 0;
      if (country == gc_data[i].gc_country)
        {
          gc_data[n].n |= 1024;
        }
    }

  /****  evaluate name  ****/
  k = 0;
  while (temp[k] != '\0'  &&  temp[k] != ' '
  &&  temp[k] != '-'  &&  temp[k] != '.')
    {
      k++;
    }
  if (temp[k] == '.')
    {
      k++;
    }

  if (temp[k] != '\0')
    {
     /****  search whole name  ****/
     gender = internal_search (temp, compare_mode, country TSRMLS_CC);

     if (gender != NAME_NOT_FOUND)
       {
         return (gender);
       }
    }

  /****  evaluate "multiple" name  ****/
  i = 0;
  gender = NAME_NOT_FOUND;
  compare_mode |= SEARCH_FIRST;
  res = 0;

  while (temp[i] != '\0')
    {
      /****  look for end of word  ****/
      k = i;
      while (temp[k] != '\0'  &&  temp[k] != ' '
      &&  temp[k] != '-'  &&  temp[k] != '.')
        {
          k++;
        }
      if (temp[k] == '.')
        {
          k++;
        }
      c = temp[k];
      temp[k] = '\0';

      if (k > i)
        {
          /****  check this name  ****/
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              php_printf ("\n");
            }
          if ((GENDER_G(internal_mode) & GENDER_GET_COUNTRY)
          &&  ! (compare_mode & SEARCH_FIRST))
            {
              for (n=0; gc_data[n].country_text != NULL; n++)
                {
                  gc_data[n].n &= ~1024;
                  gc_data[n].n *= 16;

                  if (country == gc_data[i].gc_country)
                    {
                      gc_data[n].n |= 1024;
                    }

                }
            }

          gender = internal_search (temp+i, compare_mode, country TSRMLS_CC);
          if (GENDER_G(internal_mode) & TRACE_GENDER)
            {
              trace_info ("result for", temp+i, NULL, gender, NULL TSRMLS_CC);
            }

          if ((GENDER_G(internal_mode) & GENDER_GET_COUNTRY)
          &&  ! (compare_mode & SEARCH_FIRST))
            {
              for (n=0; gc_data[n].country_text != NULL; n++)
                {
                  i = (int) ((gc_data[n].n & ~1024) / 16);
                  gc_data[n].n = multiply_frequency (i, gc_data[n].n % 16);

                  if (country == gc_data[i].gc_country)
                    {
                      gc_data[n].n |= 1024;
                    }

                }
            }
          compare_mode &= ~SEARCH_FIRST;

          switch (gender)
            {
              case IS_FEMALE        :  res |=  1;  break;
              case IS_MOSTLY_FEMALE :  res |=  2;  break;
              case IS_MALE          :  res |=  4;  break;
              case IS_MOSTLY_MALE   :  res |=  8;  break;
              case IS_UNISEX_NAME   :  res |= 16;  break;
              case NAME_NOT_FOUND   :  res |= 32;  break;
              default               :  res |= 64;  break;
            }
        }

      temp[k] = c;
      i = k;
      /****  look for begin of word  ****/
      while (temp[i] == ' '  ||  temp[i] == '-'  ||  temp[i] == '.')
        {
          i++;
        }
    }

  if ((res & 1)  &&  (res & 4))
    {
      if (compare_mode & GENDER_ALLOW_COUPLE)
        {
          /****  name is both male and female  ****/
          return (IS_A_COUPLE);
        }

      /****  error in mame  ****/
      return (ERROR_IN_NAME);
    }

  if (res & 1)
    {
      return (IS_FEMALE);
    }
  if (res & 4)
    {
      return (IS_MALE);
    }

  if ((res & 2)  &&  ! (res & 8))
    {
      return (IS_MOSTLY_FEMALE);
    }
  if ((res & 8)  &&  ! (res & 2))
    {
      return (IS_MOSTLY_MALE);
    }

  if (res & (2|8|16))
    {
      return (IS_UNISEX_NAME);
    }

  return (gender);
}




int check_nickname (char first_name_1[], char first_name_2[],
   int compare_mode, int country TSRMLS_DC)
/****  This function checks if two first names are "equivalent"  ****/

/****  Result: EQUIVALENT_NAMES :  names are equivalent  ****/
/****          NAME_NOT_FOUND   :  names not found       ****/
/****         INTERNAL_ERROR_GENDER :  internal error    ****/
{
  int  i,k,n;
  char a_temp [LENGTH_FIRST_NAME+3];
  char a2_temp [LENGTH_FIRST_NAME+3];

  i = copycut (a_temp, first_name_1, LENGTH_FIRST_NAME+3 TSRMLS_CC);
  k = copycut (a2_temp,first_name_2, LENGTH_FIRST_NAME+3 TSRMLS_CC);

  if (i > LENGTH_FIRST_NAME  ||  k > LENGTH_FIRST_NAME)
    {
      /****  name is too long  ****/
      return (ERROR_IN_NAME);
    }

  if (strcmp (a_temp,a2_temp) == 0)
    {
      /****  names are exactly the same  ****/
      return (EQUIVALENT_NAMES);
    }
  if (i == 0  ||  k == 0)
    {
      /****  at least one name is empty  ****/
      return (NOT_EQUAL_NAMES);
    }

  if (strchr (a_temp, ' ') != NULL
  ||  strchr (a_temp, '-') != NULL
  ||  strchr (a2_temp,' ') != NULL
  ||  strchr (a2_temp,'-') != NULL)
    {
      /****  at least one "multiple" name  ****/
      return (NOT_EQUAL_NAMES);
    }

  n = strcmp_search (a_temp, a2_temp, COMPARE_ABBREVIATION,0, NULL TSRMLS_CC);
  if (n == 0)
    {
      /****  names are like (e.g.) "Carl" and "C."  ****/
      return (EQUIVALENT_NAMES);
    }

  /****  reset statistics  ****/
  for (n=0; gc_data[n].country_text != NULL; n++)
    {
      gc_data[n].n = 0;
      if (country == gc_data[i].gc_country)
        {
          gc_data[n].n |= 1024;
        }

    }

  compare_mode &= ~(SEARCH_GENDER | SEARCH_FIRST);
  compare_mode |= SEARCH_EQUIV;

  n = NAME_NOT_FOUND;
  if (i <= k)
    {
      sphp_printf (a_temp+i, " %s", a2_temp);
      n = internal_search (a_temp, compare_mode, country TSRMLS_CC);
      a_temp[i] = '\0';
    }
  if (k <= i  &&  n == NAME_NOT_FOUND)
    {
      sphp_printf (a2_temp+k, " %s", a_temp);
      n = internal_search (a2_temp, compare_mode, country TSRMLS_CC);
      a2_temp[k] = '\0';
    }

  return (n);
}


void cleanup_gender (TSRMLS_D)
{
  if (GENDER_G(f_names) != NULL)
    {
      php_stream_close(GENDER_G(f_names));
      GENDER_G(f_names) = NULL;
    }
  GENDER_G(line_size) = 0;
  GENDER_G(record_count) = 0L;
}


int determine_country (char *text TSRMLS_DC)
{
  char *s;
  int  i,k,country;

  country = GC_ANY_COUNTRY;
  k = (int)strlen (text);

  if (isdigit (*text))
    {
      country = atoi (text);
    }
  for (i=0; gc_data[i].country_text != NULL; i++)
    {
      if (strcmp_search (text, gc_data[i].country_text, 0,0, NULL TSRMLS_CC) == 0
      || ((s = strstr (gc_data[i].country_text, text)) != NULL
       &&  *text == GENDER_G(upperchar)[(unsigned char) *s]
       && (*(s+k) == '\0'  ||  *(s+k) == '/'  ||  *(s+k) == ' ')))
        {
          country = gc_data[i].gc_country;
        }
    }

  return (country);
}
